import sys
import math
import pygame
import data
import const
import util
import level

class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('Walkway')
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        self.game()
    
    def game(self):
        Game(self, self.screen).loop()


class Game(object):
    
    def __init__(self, window, screen):
        self.window = window
        self.screen = screen
        
        self.clock = pygame.time.Clock()
        
        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath('Desperation.ogg'))
            pygame.mixer.music.play(-1)
        except:
            print 'Cannot load sound. :('
        
        pygame.font.init()
        self.font_18 = pygame.font.Font(data.filepath('OptimusPrincepsSemiBold.ttf'), 18)
        self.font_32 = pygame.font.Font(data.filepath('OptimusPrincepsSemiBold.ttf'), 32)
        self.font_72 = pygame.font.Font(data.filepath('OptimusPrincepsSemiBold.ttf'), 72)
        
        self.bg = pygame.image.load(data.filepath('bg.jpg')).convert()
        self.level_number = 0
        self.init_level(0)
    
    def loop(self):
        pygame.display.update()

        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)

            self.screen.blit(self.bg, (0, 0))

            dt = self.clock.tick(const.FPS) / 1000.0

            self.level.update(dt)

            if self.level_number != len(level.LEVELS) - 1:
                self.level.draw(self.screen)
                self.trail.draw(self.screen)
                self.position_con.draw(self.screen)
            
            if self.level_number >= 2:
                self.screen.blit(self.font_18.render(str(self.level_number + 1) + ' of ' + str(len(level.LEVELS) - 1), True, (255, 255, 255)), (5, 5))

            if self.level_number == 0:
                self.screen.blit(self.font_72.render('Walkway', True, (255, 255, 255)), (128, 32))
                self.screen.blit(self.font_18.render('Arrow keys to move', True, (255, 255, 255)), (190, 470))
                self.screen.blit(self.font_18.render('Z to undo', True, (255, 255, 255)), (240, 500))
            elif self.level_number == 1:
                self.screen.blit(self.font_18.render('Large nodes must be crossed', True, (255, 255, 255)), (160, 470))
                self.screen.blit(self.font_18.render('and only they allow crossing', True, (255, 255, 255)), (150, 500))
            elif self.level_number == 2:
                self.screen.blit(self.font_18.render('There are 15 levels total', True, (255, 255, 255)), (180, 500))
            elif self.level_number == 3:
                self.screen.blit(self.font_18.render('Undo (Z) is your friend!', True, (255, 255, 255)), (180, 500))
            elif self.level_number == 4:
                self.screen.blit(self.font_18.render('Good luck! It will get hard!', True, (255, 255, 255)), (160, 500))
            elif self.level_number == len(level.LEVELS) - 1:
                self.screen.blit(self.font_32.render('You\'re awesome!', True, (255, 255, 255)), (160, 200))                
                self.screen.blit(self.font_32.render('Thanks for playing!', True, (255, 255, 255)), (130, 300))                
        
            pygame.display.flip()
            
            if self.level.done:
                self.level_number += 1
                if self.level_number < len(level.LEVELS):
                    self.init_level(self.level_number)
    
    def on_keydown(self, event):
        if event.key == pygame.K_UP:
            if self.level.is_blocked((self.position.cpos[0], self.position.cpos[1] - 1)): return
            if self.trail.get_last_direction() == (0,1): return
            self.trail.add_trail(self.position.cpos, (0,-1))
            prev_pos = self.position.cpos
            self.position.move_up()
            self.level.update_pos(prev_pos, self.position.cpos)
        elif event.key == pygame.K_DOWN:
            if self.level.is_blocked((self.position.cpos[0], self.position.cpos[1] + 1)): return
            if self.trail.get_last_direction() == (0,-1): return
            self.trail.add_trail(self.position.cpos, (0,1))
            prev_pos = self.position.cpos
            self.position.move_down()
            self.level.update_pos(prev_pos, self.position.cpos)
        elif event.key == pygame.K_LEFT:
            if self.level.is_blocked((self.position.cpos[0] - 1, self.position.cpos[1])): return
            if self.trail.get_last_direction() == (1,0): return
            self.trail.add_trail(self.position.cpos, (-1,0))
            prev_pos = self.position.cpos
            self.position.move_left()
            self.level.update_pos(prev_pos, self.position.cpos)
        elif event.key == pygame.K_RIGHT:
            if self.level.is_blocked((self.position.cpos[0] + 1, self.position.cpos[1])): return
            if self.trail.get_last_direction() == (-1,0): return
            self.trail.add_trail(self.position.cpos, (1,0))
            prev_pos = self.position.cpos
            self.position.move_right()
            self.level.update_pos(prev_pos, self.position.cpos)
        elif event.key == pygame.K_z:
            last_direction = self.trail.undo()
            if last_direction is not None:
                prev_pos = self.position.cpos
                self.position.undo(last_direction)
                self.level.update_pos_undo(prev_pos, self.position.cpos)
    
    def on_keyup(self, event):
        if event.key == pygame.K_ESCAPE:
            sys.exit()
    
    def init_level(self, level_number):
        self.trail = Trail()
        self.level = Level(level.LEVELS[level_number], self.trail)
        self.level.update_nodes()
        self.position = Position(self.level.starting_point)
        self.position_con = pygame.sprite.Group()
        self.position_con.add(self.position)


class Level(object):
    
    def __init__(self, level_str, trail):
        self.items = []
        self.tilemap = [[True for j in xrange(const.NUMBER_OF_CELLS)] for i in xrange(const.NUMBER_OF_CELLS)]
        self.blocked = [[False for j in xrange(const.NUMBER_OF_CELLS)] for i in xrange(const.NUMBER_OF_CELLS)]
        self.starting_point = (7, 7)
        self.nodes = pygame.sprite.Group()
        self.gold = pygame.sprite.Group()
        self.goldmap = [[0 for j in xrange(const.NUMBER_OF_CELLS)] for i in xrange(const.NUMBER_OF_CELLS)]        
        self.total_gold = 0
        self.trail = trail

        self.read_map(level_str)
        
        self.done = False
    
    def update_nodes(self):
        for i in xrange(const.NUMBER_OF_CELLS):
            for j in xrange(const.NUMBER_OF_CELLS):
                if self.tilemap[i][j]:
                    self.nodes.add(Node((i, j)))
                else:
                    self.blocked[i][j] = True
        self.starting_point_spr = pygame.sprite.Group()
        sp_sprite = pygame.sprite.Sprite()
        sp_sprite.image = util.load_image('startingpoint.png')
        sp_sprite.rect = sp_sprite.image.get_rect()
        sp_sprite.rect.center = util.cpos_to_pos(self.starting_point)
        self.starting_point_spr.add(sp_sprite)
    
    def set_blocked(self, cpos, blocked=True):
        self.blocked[cpos[0]][cpos[1]] = blocked
    
    def is_blocked(self, cpos):
        if cpos[0] < 0 or cpos[0] >= const.NUMBER_OF_CELLS: return True
        if cpos[1] < 0 or cpos[1] >= const.NUMBER_OF_CELLS: return True
        return self.blocked[cpos[0]][cpos[1]]
    
    def add_gold(self, cpos, double=False):
        if double:
            gold = DoubleGold(cpos)
            self.goldmap[cpos[0]][cpos[1]] = 2
            self.total_gold += 2
        else:
            gold = Gold(cpos)
            self.goldmap[cpos[0]][cpos[1]] = 1
            self.total_gold += 1
        self.gold.add(gold)
    
    def update_pos(self, prev_pos, cpos):
        if self.goldmap[prev_pos[0]][prev_pos[1]] != 2:
            self.set_blocked(prev_pos)
        elif not self.trail.is_last_trail_straight():
            self.set_blocked(prev_pos)
        if self.goldmap[cpos[0]][cpos[1]] != 0:
            self.total_gold -= 1
        if self.total_gold <= 0:
            self.done = True

    def update_pos_undo(self, prev_pos, cpos):
        self.set_blocked(cpos, False)
        if self.goldmap[prev_pos[0]][prev_pos[1]] != 0:
            self.total_gold += 1
        
    def read_map(self, map_str):
        split_map = map_str.split('\n')
        if split_map[0] == '': split_map = split_map[1:]
        j = 0
        for line in split_map:
            split_line = line.split(' ')
            i = 0
            for cell in split_line:
                if cell == '.':
                    self.tilemap[i][j] = False
                elif cell == 'X':
                    self.starting_point = (i,j)
                elif cell == '1':
                    self.add_gold((i,j))
                elif cell == '2':
                    self.add_gold((i,j), True)
                i += 1
            j += 1
    
    def update(self, dt):
        pass

    def draw(self, screen):
        self.nodes.draw(screen)
        self.gold.draw(screen)
        self.starting_point_spr.draw(screen)


class Node(pygame.sprite.Sprite):
    
    def __init__(self, cpos):
        pygame.sprite.Sprite.__init__(self)
        self.cpos = list(cpos)
        self.pos = util.cpos_to_pos(cpos)
        self.image = util.load_image('node.png')
        self.rect = self.image.get_rect()
        self.rect.center = self.pos


class Position(pygame.sprite.Sprite):
    
    def __init__(self, cpos):
        pygame.sprite.Sprite.__init__(self)
        self.cpos = list(cpos)
        self.pos = util.cpos_to_pos(cpos)
        self.image = util.load_image('position.png')
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
    
    def move_up(self):
        if self.cpos[1] <= 0: return
        self.cpos = [self.cpos[0], self.cpos[1] - 1]
        self.update_pos()
    
    def move_down(self):
        if self.cpos[1] >= const.NUMBER_OF_CELLS - 1: return
        self.cpos = [self.cpos[0], self.cpos[1] + 1]
        self.update_pos()
    
    def move_left(self):
        if self.cpos[0] <= 0: return
        self.cpos = [self.cpos[0] - 1, self.cpos[1]]
        self.update_pos()
    
    def move_right(self):
        if self.cpos[0] >= const.NUMBER_OF_CELLS - 1: return
        self.cpos = [self.cpos[0] + 1, self.cpos[1]]
        self.update_pos()
    
    def update_pos(self):
        self.pos = util.cpos_to_pos(self.cpos)
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
    
    def undo(self, last_direction):
        if last_direction[0] < 0:
            self.move_right()
        if last_direction[0] > 0:
            self.move_left()
        if last_direction[1] < 0:
            self.move_down()
        if last_direction[1] > 0:
            self.move_up()


class Trail(pygame.sprite.Group):
    
    def __init__(self):
        pygame.sprite.Group.__init__(self)
        self.trails = []
    
    def add_trail(self, from_pos, direction):
        trail = TrailSegment(from_pos, direction)
        self.add(trail)
        self.trails.append(trail)
    
    def undo(self):
        if len(self.trails) == 0: return None
        last_trail = self.trails[-1]
        self.remove(last_trail)
        self.trails.pop()
        return last_trail.direction
    
    def is_last_trail_straight(self):
        if len(self.trails) == 1: return False
        return self.trails[-2].direction == self.trails[-1].direction
    
    def get_last_direction(self):
        if len(self.trails) == 0: return (0, 0)
        return self.trails[-1].direction


class TrailSegment(pygame.sprite.Sprite):

    def __init__(self, from_pos, direction):
        pygame.sprite.Sprite.__init__(self)
        self.pos = util.cpos_to_pos(from_pos)
        self.direction = direction
        if direction[0] != 0:
            self.image = pygame.surface.Surface((const.CELL_SIZE+2,4))
        elif direction[1] != 0:
            self.image = pygame.surface.Surface((4,const.CELL_SIZE+2))
        self.image.fill((200,50,50))
        self.rect = self.image.get_rect()
        if direction[0] < 0:
            self.pos = [self.pos[0] + 1, self.pos[1]]
            self.rect.midright = self.pos
        elif direction[0] > 0:
            self.pos = [self.pos[0] - 1, self.pos[1]]
            self.rect.midleft = self.pos
        elif direction[1] < 0:
            self.pos = [self.pos[0], self.pos[1] + 1]
            self.rect.midbottom = self.pos
        elif direction[1] > 0:
            self.pos = [self.pos[0], self.pos[1] - 1]
            self.rect.midtop = self.pos

        
class Gold(pygame.sprite.Sprite):
    
    def __init__(self, cpos):
        pygame.sprite.Sprite.__init__(self)
        self.cpos = list(cpos)
        self.pos = util.cpos_to_pos(cpos)
        self.image = util.load_image('gold.png')
        self.rect = self.image.get_rect()
        self.rect.center = self.pos


class DoubleGold(pygame.sprite.Sprite):

    def __init__(self, cpos):
        pygame.sprite.Sprite.__init__(self)
        self.cpos = list(cpos)
        self.pos = util.cpos_to_pos(cpos)
        self.image = util.load_image('doublegold.png')
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
